# MIDI SCRIPTER
# Copyright (c) 2005 Emanuel Borsboom.  All rights reserved.
#
# UnsafeProxy class, which allows you to call code that is running at a lower $SAFE level.
#

require 'thread'

class UnsafeProxy

	def initialize(obj, *methods)
		@obj = obj
		@methods = methods
		if (@methods.length == 0)
			@methods = obj.class.public_instance_methods(false)
			@methods.collect! { |x| x.intern }
		end

		thr = Thread.new do
			# Create instance variables at safe level 3 so they are tainted so that safe level 4 code can use them
			$SAFE = 3
			@meth = nil
			@args = nil
			@block = nil
			@result = nil
			@exception = nil
			@mutex = Mutex.new
			@call_cv = ConditionVariable.new
			@result_cv = ConditionVariable.new
		end
		thr.join
		self.taint # Taint self so that safe level 4 code can use me
		
		Thread.new do
			# Start the thread that delegates calls to the real class.  This stays at the original safe level (probably 0)
			@mutex.synchronize do
				while true
					@call_cv.wait(@mutex)
					@result = nil
					@exception = nil
					begin
						if !@methods.include? @meth
							raise SecurityError, "No permission to call method '" + @meth.to_s + "'", caller
						end
						if @block != nil
							@result = @obj.send(@meth, *@args) { @block.call }
						else
							@result = @obj.send(@meth, *@args)
						end
					rescue => excep
						#puts "EXCEPTION: " + excep.to_s + "\n" + excep.backtrace.join("\n")
						@exception = excep
					end
					@result_cv.signal
				end
			end
		end
	end
	
	def method_missing(meth, *args, &block)
		@mutex.synchronize do
			@meth = meth
			@args = args			
			@block = block
			@call_cv.signal
			@result_cv.wait(@mutex)
			if @exception != nil
				raise @exception, @exception.to_s, caller[2..-1]
			end
			@result
		end	
	end

end


#~ class Printer
	#~ def print(s)
		#~ puts s
	#~ end	
	#~ def get_answer
		#~ 42
	#~ end
	#~ def run_block
		#~ 42 + yield
	#~ end
#~ end

#~ thr = Thread.new do
	#~ $printer = UnsafeProxy.new(Printer.new)
	#~ $SAFE = 4
	#~ Thread.critical = true
	#~ Thread.critical = false
	#~ begin
		#~ $printer.print "hello, world"
		#~ $printer.print "answer: " + $printer.get_answer.to_s
		#~ v = $printer.run_block do
			#~ #system "dir"
			#~ 42
		#~ end
		#~ $printer.print "run_block: " + v.to_s
		#~ $printer.system "dir"
	#~ rescue
		#~ $printer.print "EXCEPTION: " + $!.to_s + "\n" + $!.backtrace.join("\n")
	#~ end
#~ end
#~ thr.join
